#!/usr/bin/env python
# -*- coding: utf8 -*-
import numpy as np
import itertools
from collections import Counter

# Ground truth is usually stored in ARFF files with a separate column for each hand-labelling expert (if multiple are
# present). Each hand-labelling expert column contains numerical data with the following correspondence to eye movement
# types:
CORRESPONDENCE_TO_HAND_LABELLING_VALUES = {
    'UNKNOWN': 0,
    'FIX': 1,
    'SACCADE': 2,
    'SP': 3,
    'NOISE': 4
}


def get_majority_vote_efficient(obj, experts, positive_label):
    """
    Get majority vote of labels among different experts, in a more efficient way through numpy functions.

    :param obj: arff object
    :param experts: list of experts (i.e. column or attribute names that are considered in the vote).
    :param positive_label: the label to be evaluated (all values in @experts columns are treated as "yes"
                           (i.e. equals to the positive label) or "no" (i.e. does not)).
    :return: majority vote of label as binary values, with 1 meaning that the majority agrees to assign
             the @positive_label to this row of @obj['data'], and 0 -- that it does not.

    Example:
    get_majority_vote_efficient(arff_object, ['expert1', 'expert2'], 'SP')

    """
    assert len(experts) >= 1
    if len(experts) == 1:
        # just one expert, he always agrees with himself
        return obj['data'][experts[0]] == CORRESPONDENCE_TO_HAND_LABELLING_VALUES[positive_label]
    else:
        hand_labellings = obj['data'][experts]
        hand_labellings_list = hand_labellings.view((np.float32, len(hand_labellings.dtype.names)))
        thd = len(experts) / 2.0  # at least 50% of experts should agree
        majority_vote = ((hand_labellings_list == CORRESPONDENCE_TO_HAND_LABELLING_VALUES[positive_label]).sum(axis=1)
                         >= thd).astype(int)
        return majority_vote


def get_majority_vote(obj, experts, exclude_values=None):
    """
    Get majority vote of labels among different experts, keeping all the labels as possible.

    :param obj: arff objects
    :param experts: list of experts (i.e. column or attribute names that are considered in the vote).
    :param exclude_values: a list of values that should be excluded from the vote (ex. ['UNKNOWN'], or [0]);
                           if not a list, will be converted to a list with 1 element;
                           if all the experts vote for one of these values, they will be taken into account.
    :return: majority vote of label in form of array of label number.

    Example:
    get_majority_vote(arff_object1, ['expert1', 'expert2'])

    """
    if len(experts) == 1:
        return obj['data'][experts].astype(int)
    else:
        if exclude_values is not None:
            if not isinstance(exclude_values, list):
                exclude_values = [exclude_values]
        else:
            exclude_values = []
        exclude_values = set(exclude_values)

        majority_vote = []
        hand_labellings = obj['data'][experts].tolist()
        for i in xrange(len(hand_labellings)):
            # max number of occurrences wins
            candidates_set = set(hand_labellings[i]).difference(exclude_values)
            if not candidates_set:
                candidates_set = set(hand_labellings[i])
            majority_vote.append(max(candidates_set, key=hand_labellings[i].count))
        majority_vote = np.asarray(majority_vote, dtype=np.int)

        return majority_vote


def compute_statistics(raw_stats):
    """
    Convert raw statistics in @raw_stats to some more usable format
    :param raw_stats: a dictionary with keys "TP", "TN", "FP", "FN" for true positives count, true negatives,
                      false positives and false negatives, respectively.
    :return: a dictionary with keys "precision", "recall", "F1", "FPR" (for False Positive Rate), "accuracy".
    """

    # for Kappa computation
    if (raw_stats['TP'] + raw_stats['FP'] + raw_stats['TN'] + raw_stats['FN']) != 0:
        p_chance = float((raw_stats['TP'] + raw_stats['FP']) * (raw_stats['TP'] + raw_stats['FN']) +
                         (raw_stats['FP'] + raw_stats['TN']) * (raw_stats['FN'] + raw_stats['TN'])) / \
                   ((raw_stats['TP'] + raw_stats['FP'] + raw_stats['TN'] + raw_stats['FN']) ** 2)
        p_observed = float(raw_stats['TP'] + raw_stats['TN']) / (raw_stats['TP'] +
                                                                 raw_stats['FP'] +
                                                                 raw_stats['TN'] +
                                                                 raw_stats['FN'])
    else:
        p_chance = 0.0
        p_observed = 0.0

    res = {
        'precision': (float(raw_stats['TP']) / (raw_stats['TP'] + raw_stats['FP']))
        if (raw_stats['TP'] + raw_stats['FP']) != 0 else 0.0,

        'recall': (float(raw_stats['TP']) / (raw_stats['TP'] + raw_stats['FN']))
        if (raw_stats['TP'] + raw_stats['FN']) != 0 else 0.0,

        'F1': (2 * float(raw_stats['TP']) / (2 * raw_stats['TP'] + raw_stats['FP'] + raw_stats['FN']))
        if (2 * raw_stats['TP'] + raw_stats['FP'] + raw_stats['FN']) != 0 else 0.0,

        'FPR': (float(raw_stats['FP']) / (raw_stats['FP'] + raw_stats['TN']))
        if (raw_stats['FP'] + raw_stats['TN']) != 0 else 0.0,

        'accuracy':
            (float(raw_stats['TP'] + raw_stats['TN']) / (raw_stats['TP'] +
                                                         raw_stats['FP'] +
                                                         raw_stats['TN'] +
                                                         raw_stats['FN']))
        if (raw_stats['TP'] + raw_stats['FP'] + raw_stats['TN'] + raw_stats['FN']) != 0 else 0.0,

        'sensitivity': (float(raw_stats['TP']) / (raw_stats['TP'] + raw_stats['FN']))
        if (raw_stats['TP'] + raw_stats['FN']) != 0 else 0.0,

        'specificity': (float(raw_stats['TN']) / (raw_stats['TN'] + raw_stats['FP']))
        if (raw_stats['TN'] + raw_stats['FP']) != 0 else 0.0,

        'kappa': ((p_observed - p_chance) / (1 - p_chance)) if p_chance != 1.0 else 0.0
    }
    return res


def evaluate_episodes(true_labels_list, assigned_labels_list, experts, positive_label='SP', return_raw_stats=False,
                      interval_vs_interval=False):
    """
    Event-level evaluation of labelling result (algorithm output) in @assigned_labels_list with
    hand-labelling expert's labels in @ground_truth_list.

    :param true_labels_list: list of arff objects produced with hand-labelling tool [1].
    :param assigned_labels_list: list of arff objects produced with this tool (or loaded via RecordingProcessor).
    :param experts: list of experts (for our data, one expert was the tie-corrector, so normally a list of one element
                    should be used).
    :param positive_label: the positive abel to be evaluated (normally would be 'SP'/'FIX'/'SACCADE').
    :param return_raw_stats: whether to return raw statistics (a confusion matrix) or nicer (per-class accuracy) ones
    :param interval_vs_interval: if False, will check if EM episodes in the ground truth are covered by at least 50%
                                 of correct labels;
                                 if True, will check if at least 50% of those episodes is covered by a continuous
                                 interval of the correct label.
    :return: evaluation results in a dictionary form

    [1] http://ieeexplore.ieee.org/abstract/document/7851169/
    """
    labels = ['FIX', 'SACCADE', 'SP', 'NOISE']  # skip the UNKNOWN label
    raw_confusion = {k: 0.0 for k in labels}  # count the number of hits for each class
    raw_confusion_denominator = 0.0

    raw_stats = {
        'TP': 0.0,
        'FP': 0.0,
        'TN': 0.0,
        'FN': 0.0
    }

    for ground_truth, assigned_labels in zip(true_labels_list, assigned_labels_list):
        # check that the t-x-y data has all at least similar values
        assert np.allclose(ground_truth['data']['time'], assigned_labels['data']['time'])
        assert np.allclose(ground_truth['data']['x'], assigned_labels['data']['x'])
        assert np.allclose(ground_truth['data']['y'], assigned_labels['data']['y'])

        ground_truth_labels = get_majority_vote(ground_truth, experts)

        start_i = 0
        for current_label, grp in itertools.groupby(ground_truth_labels):
            grp_len = len(list(grp))

            alg_labels = assigned_labels['data']['EYE_MOVEMENT_TYPE'][start_i:start_i + grp_len]
            for equivalent_noise_label in ['NOISE_CLUSTER', 'BLINK']:
                alg_labels[alg_labels == equivalent_noise_label] = 'NOISE'

            if not interval_vs_interval:  # check if at least 50% of the interval is covered by correct labels
                alg_labels = Counter(alg_labels)
                alg_majority_label, alg_majority_size = alg_labels.most_common(1)[0]
                if alg_majority_size >= 0.5 * grp_len:
                    # a hit for the current label in the @alg_majority_label column
                    pass
                else:
                    alg_majority_label = 'UNKNOWN'  # no label being assigned
            else:  # check if at least 50% of the interval is covered by a continuous interval of correct labels
                label_groups = itertools.groupby(alg_labels)
                label_groups = [(k, len(list(v))) for k, v in label_groups]
                label_values = [k for k, _ in label_groups]
                label_counts = [v for _, v in label_groups]

                max_i = np.argmax(label_counts)
                if label_counts[max_i] >= 0.5 * grp_len:
                    alg_majority_label = label_values[max_i]
                else:
                    alg_majority_label = 'UNKNOWN'

            # Record confusion matrix row and the @raw_stats.
            # Ensure that the @current_label - 1 is in tha valid range
            # (otherwise, it is some extra label, like PSO, which we ignore).
            if 0 <= current_label - 1 < len(labels) and labels[current_label - 1] == positive_label:
                raw_confusion_denominator += 1
                if alg_majority_label in labels:
                    raw_confusion[alg_majority_label] += 1

                if alg_majority_label == positive_label:
                    # true: +, detected: +
                    raw_stats['TP'] += 1
                else:
                    # true: +, detected: -
                    raw_stats['FN'] += 1
            else:
                if alg_majority_label == positive_label:
                    # true: -, detected: +
                    raw_stats['FP'] += 1
                else:
                    # true: -, detected: -
                    raw_stats['TN'] += 1

            start_i += grp_len

    raw_confusion = {k: (raw_confusion[k] / raw_confusion_denominator) if raw_confusion_denominator != 0 else 0.0 for k in labels}

    if return_raw_stats:
        raw_stats['confusion'] = raw_confusion
        return raw_stats
    else:
        stats = compute_statistics(raw_stats)
        stats['confusion'] = raw_confusion
        stats['confusion-accuracy'] = raw_confusion[positive_label]
        return stats


def evaluate(true_labels_list, assigned_labels_list, experts, positive_label='SP', return_raw_stats=False):
    """
    Evaluate labelling result (algorithm output) in @assigned_labels_list with hand-labelling expert's labels
    in @ground_truth_list.

    :param true_labels_list: list of arff objects produced with hand-labelling tool [1].
    :param assigned_labels_list: list of arff objects produced with this tool (or loaded via RecordingProcessor).
    :param experts: list of experts (for our data, one expert was the tie-corrector, so normally a list of one element
                    should be used).
    :param positive_label: the positive abel to be evaluated (normally would be 'SP'/'FIX'/'SACCADE').
    :param return_raw_stats: whether to return raw statistics (TP/TN/FP/FN) or nicer (F1/precision/recall/...) ones
    :return: evaluation results in a dictionary form

    [1] http://ieeexplore.ieee.org/abstract/document/7851169/
    """
    raw_stats = {
        'TP': 0,
        'FP': 0,
        'TN': 0,
        'FN': 0
    }
    for ground_truth, assigned_labels in zip(true_labels_list, assigned_labels_list):
        # check that the t-x-y data has all at least similar values
        assert np.allclose(ground_truth['data']['time'], assigned_labels['data']['time'])
        assert np.allclose(ground_truth['data']['x'], assigned_labels['data']['x'])
        assert np.allclose(ground_truth['data']['y'], assigned_labels['data']['y'])

        assigned_labels_status_list = (assigned_labels['data']['EYE_MOVEMENT_TYPE'] == positive_label).astype(int)

        ground_truth_status_list = get_majority_vote_efficient(ground_truth, experts, positive_label)
        # ground_truth_status_list = (get_majority_vote(ground_truth, experts) ==
        #                             CORRESPONDENCE_TO_HAND_LABELLING_VALUES[positive_label]).astype(int)

        raw_stats['TP'] += ((ground_truth_status_list == 1) * (assigned_labels_status_list == 1)).sum()
        raw_stats['FP'] += ((ground_truth_status_list == 0) * (assigned_labels_status_list == 1)).sum()
        raw_stats['TN'] += ((ground_truth_status_list == 0) * (assigned_labels_status_list == 0)).sum()
        raw_stats['FN'] += ((ground_truth_status_list == 1) * (assigned_labels_status_list == 0)).sum()

    if return_raw_stats:
        stats = raw_stats
    else:
        stats = compute_statistics(raw_stats)

    stats['episode'] = evaluate_episodes(true_labels_list=true_labels_list,
                                         assigned_labels_list=assigned_labels_list,
                                         experts=experts,
                                         positive_label=positive_label,
                                         return_raw_stats=return_raw_stats,
                                         interval_vs_interval=False)

    stats['episode-vs-episode'] = evaluate_episodes(true_labels_list=true_labels_list,
                                                    assigned_labels_list=assigned_labels_list,
                                                    experts=experts,
                                                    positive_label=positive_label,
                                                    return_raw_stats=return_raw_stats,
                                                    interval_vs_interval=True)
    return stats
